from pathlib import Path


class ScriptLocation(object):
    def __init__(self, file_name: str = "<unknown>", function_name: str = "<unknown>", line: str = "<unknown>"):
        self.file_name = file_name if file_name is not None else "<unknown>"
        self.function_name = function_name if function_name is not None else "<unknown>"
        self.line = str(line) if line is not None else "<unknown>"

    def get_int_line(self):
        try:
            return int(self.line)
        except Exception:
            return 0

    def __str__(self):
        return f"{self.file_name}:{self.line} ({self.function_name})"

    def __repr__(self):
        return str(self)

    @staticmethod
    def parse_from_dict(data: dict):

        file_name = Path(data.get("file_name", "unknown")).name
        function_name = data.get("function_name", "unknown")
        line = str(data.get("line", -1))

        if 'fileName' in data.keys():
            file_name = Path(data.get("fileName", file_name)).name

        if 'lineNumber' in data.keys():
            line = str(data.get("lineNumber", line))

        return ScriptLocation(
            file_name=file_name,
            function_name=function_name,
            line=line
        )



import json
import os.path
from pathlib import Path
import base64
import string

from frida_fusion.libs.logger import Logger
from frida_fusion.libs.database import Database
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class Settings(ModuleBase):
    class SettingsDB(Database):
        dbName = ""

        def __init__(self, db_name: str):
            super().__init__(
                auto_create=True,
                db_name=db_name
            )
            self.create_db()

        def create_db(self):
            super().create_db()
            conn = self.connect_to_db(check=False)

            # definindo um cursor
            cursor = conn.cursor()

            # criando a tabela (schema)
            cursor.execute("""
                CREATE TABLE IF NOT EXISTS [android_settings] (
                    id INTEGER PRIMARY KEY AUTOINCREMENT,
                    module TEXT NULL,
                    name TEXT NULL,
                    flag INTEGER NULL DEFAULT (0),
                    data TEXT NULL,
                    created_date datetime not null DEFAULT (datetime('now','localtime')),
                    UNIQUE (module, name, flag, data)
                );
            """)

            conn.commit()

            # Must get the constraints
            self.get_constraints(conn)

    def __init__(self):
        super().__init__('Settings', 'Hook Android Settings functions')
        self._settings_db = None
        self.mod_path = str(Path(__file__).resolve().parent)

    def start_module(self, **kwargs) -> bool:
        if 'db_path' not in kwargs:
            raise Exception("parameter db_path not found")

        self._settings_db = Settings.SettingsDB(db_name=kwargs['db_path'])
        return True

    def js_files(self) -> list:
        return [
            os.path.join(self.mod_path, "settings.js")
        ]

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:

        if module in ["Settings$Secure.getString",
                      "Settings$Secure.putString",
                      "Settings$Secure.getUriFor",
                      "Settings$Secure.getInt",
                      "Settings$System.getString",
                      "Settings$System.putString",
                      "Settings$System.getUriFor",
                      "Settings$System.getInt",
                      "Settings$Global.getInt"
                      ]:
            name = received_data.get('name', None)
            flag = received_data.get('flag', 0)
            value = received_data.get('value', None)
            result = received_data.get('result', value)

            self._settings_db.insert_ignore_one(
                table_name='android_settings',
                module=module,
                name=name,
                flag=flag,
                data=result
            )

        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        #Nothing by now
        return True





#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import math
import shutil
import os.path
import sqlite3
from functools import reduce
from pathlib import Path
from sqlite3 import Connection, ProgrammingError


# TODO: use this decorator to wrap commit/rollback in a try/except block ?
# see http://www.kylev.com/2009/05/22/python-decorators-and-database-idioms/
from frida_fusion.libs.color import Color


def connect(func):
    """Decorator to (re)open a sqlite database connection when needed.

    A database connection must be open when we want to perform a database query
    but we are in one of the following situations:
    1) there is no connection
    2) the connection is closed

    Parameters
    ----------
    func : function
        function which performs the database query

    Returns
    -------
    inner func : function
    """

    def inner_func(self, *args, **kwargs):
        if f'{func.__module__}.{func.__qualname__}' != f'{Database.__module__}.{Database.__qualname__}.{func.__name__}':
            raise Exception('The connect decorator cannot be used outside of Database class')

        if not isinstance(self, Database):
            raise Exception('The connect decorator cannot be used outside of Database class')

        conn = kwargs.get('conn', None) if kwargs is not None else None
        if conn is None:
            conn = self.connect_to_db()

        return func(self, conn, *args, **kwargs)

    return inner_func


class Database(object):
    db_name = ""

    # Static value
    constraints = []

    def __init__(self, auto_create: bool = True, db_name: str = "fusion.db"):

        self.db_name = str(Path(db_name).resolve())

        if not os.path.isfile(self.db_name):
            if auto_create:
                self.create_db()
            else:
                raise Exception("Database not found")
        else:
            self.connect_to_db()

    def has_data(self) -> bool:
        return self.select_count('history') > 0

    def check_open(self) -> bool:
        return self.select_count('history') >= 0

    @connect
    def insert_one(self, conn: Connection, table_name, **kwargs):
        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)
        sql = "INSERT INTO {} ({}) VALUES ({})" \
            .format(table_name, ','.join(columns), ', '.join(['?'] * len(columns)))
        conn.execute(sql, values)
        conn.commit()

    @connect
    def insert_if_not_exists(self, conn: Connection, table_name, **kwargs):
        table_name = self.scrub(table_name)

        if self.select_count(table_name=table_name, **kwargs) == 0:
            (columns, values) = self.parse_args(kwargs)
            sql = "INSERT INTO {} ({}) VALUES ({})" \
                .format(table_name, ','.join(columns), ', '.join(['?'] * len(columns)))
            conn.execute(sql, values)
            conn.commit()

    @connect
    def insert_ignore_one(self, conn: Connection, table_name, **kwargs):
        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)
        sql = "INSERT OR IGNORE INTO {} ({}) VALUES ({})" \
            .format(table_name, ','.join(columns), ', '.join(['?'] * len(columns)))
        conn.execute(sql, values)
        conn.commit()

    @connect
    def insert_replace_one(self, conn: Connection, table_name, **kwargs):
        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)
        sql = "INSERT OR REPLACE INTO {} ({}) VALUES ({})" \
            .format(table_name, ','.join(columns), ', '.join(['?'] * len(columns)))
        conn.execute(sql, values)
        conn.commit()

    def insert_update_one(self, table_name: str, **kwargs):
        return self.insert_update_one_exclude(table_name, [], **kwargs)

    @connect
    def insert_update_one_exclude(self, conn: Connection, table_name: str, exclude_on_update: list = [], **kwargs) -> dict:
        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)
        sql = "INSERT OR IGNORE INTO {} ({}) VALUES ({})" \
            .format(table_name, ','.join(columns), ', '.join(['?'] * len(columns)))
        c = conn.execute(sql, values)

        status = {'inserted': c.rowcount, 'updated': 0}

        # No inserted, need to update
        if c.rowcount == 0:
            table_name = self.scrub(table_name)
            f_columns = self.constraints[table_name]
            f_values = tuple([kwargs.get(c, None) for c in f_columns],)
            args = {k: v for k, v in kwargs.items() if k not in exclude_on_update}
            (u_columns, u_values) = self.parse_args(args)

            sql = f"UPDATE {table_name} SET "
            sql += "{}".format(', '.join([f'{col} = ?' for col in u_columns]))
            if len(f_columns) > 0:
                sql += " WHERE {}".format(f' and '.join([f'{col} = ?' for col in f_columns]))
            c = conn.execute(sql, tuple(u_values + f_values, ))
            conn.commit()

            status['updated'] = c.rowcount

        conn.commit()
        return status

    @connect
    def select(self, conn: Connection, table_name, **kwargs):

        operator = self.scrub(kwargs.get('__operator', 'and'))

        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)

        sql = f"SELECT * FROM {table_name}"
        if len(columns) > 0:
            sql += " WHERE {}".format(f' {operator} '.join([f'{col} = ?' for col in columns]))

        cursor = conn.execute(sql, values)
        if cursor.rowcount == 0:
            return []

        columns = cursor.description
        return [{columns[index][0]: column for index, column in enumerate(value)} for value in cursor.fetchall()]

    def select_first(self, table_name, **kwargs):
        data = self.select(table_name, **kwargs)
        if len(data) == 0:
            return None
        return data[0]

    @connect
    def select_raw(self, conn: Connection, sql: str, args: any):
        cursor = conn.execute(sql, tuple(args,))
        if cursor.rowcount == 0:
            return []
        columns = cursor.description
        return [{columns[index][0]: column for index, column in enumerate(value)} for value in cursor.fetchall()]

    @connect
    def select_count(self, conn: Connection, table_name, **kwargs) -> int:

        operator = self.scrub(kwargs.get('__operator', 'and'))

        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)

        sql = f"SELECT count(*) FROM {table_name}"
        if len(columns) > 0:
            sql += " WHERE {}".format(f' {operator} '.join([f'{col} = ?' for col in columns]))
        cursor = conn.execute(sql, values)
        if cursor.rowcount == 0:
            return 0
        data = cursor.fetchone()

        return int(data[0])

    @connect
    def delete(self, conn: Connection, table_name, **kwargs) -> None:

        operator = self.scrub(kwargs.get('__operator', 'and'))

        table_name = self.scrub(table_name)
        (columns, values) = self.parse_args(kwargs)

        sql = f"DELETE FROM {table_name}"
        if len(columns) > 0:
            sql += " WHERE {}".format(f' {operator} '.join([f'{col} = ?' for col in columns]))
        conn.execute(sql, values)
        conn.commit()

    @connect
    def update(self, conn: Connection, table_name, filter_data, **kwargs):

        operator = self.scrub(kwargs.get('__operator', 'and'))

        table_name = self.scrub(table_name)
        (f_columns, f_values) = self.parse_args(filter_data)
        (u_columns, u_values) = self.parse_args(kwargs)

        sql = f"UPDATE {table_name} SET "
        sql += "{}".format(', '.join([f'{col} = ?' for col in u_columns]))
        if len(f_columns) > 0:
            sql += " WHERE {}".format(f' {operator} '.join([f'{col} = ?' for col in f_columns]))
        conn.execute(sql, tuple(u_values + f_values, ))
        conn.commit()

    def get_constraints(self, conn: Connection) -> dict:
        sql = ('SELECT '
               '  m.tbl_name AS table_name, '
               '  il.name AS key_name, '
               '  ii.name AS column_name '
               'FROM  '
               '  sqlite_master AS m,  '
               '  pragma_index_list(m.name) AS il,  '
               '  pragma_index_info(il.name) AS ii  '
               'WHERE  '
               '  m.type = "table" AND  '
               '  il.origin = "u"  '
               'ORDER BY table_name, key_name, ii.seqno')

        cursor = conn.execute(sql)
        columns = cursor.description
        db_scheme = [{columns[index][0]: column for index, column in enumerate(value)} for value in cursor.fetchall()]

        if len(db_scheme) > 0:
            self.constraints = reduce(lambda a, b: {**a, **b},
                                      [{table: [
                                            v['column_name'] for idx, v in enumerate(db_scheme)
                                            if v['table_name'] == table
                                        ]} for table in set([t['table_name'] for t in db_scheme])])
        else:
            self.constraints = {}

        return self.constraints

    def parse_args(self, source_dict) -> tuple:
        if source_dict is None:
            return [], tuple([])

        if not isinstance(source_dict, dict):
            raise Exception('kwargs is not a dictionary')

        columns = []
        values = []

        for key, value in source_dict.items():
            try:
                if key[0:2] != '__':
                    columns.append(f"[{self.scrub(key)}]")
                    values.append(value)
            except Exception as e:
                raise Exception(f'Error parsing {key}: {value}', e)

        return columns, tuple(values, )

    def connect_to_db(self, check: bool = True) -> Connection:
        """Connect to a sqlite DB. Create the database if there isn't one yet.

        Open a connection to a SQLite DB (either a DB file or an in-memory DB).
        When a database is accessed by multiple connections, and one of the
        processes modifies the database, the SQLite database is locked until that
        transaction is committed.

        Returns
        -------
        connection : sqlite3.Connection
            connection object
        """

        conn = sqlite3.connect(self.db_name, detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES)
        conn.create_function('log', 2, math.log)

        if check:
            try:
                # I don't know if this is the simplest and fastest query to try
                conn.execute(
                    'SELECT name FROM sqlite_temp_master WHERE type="table";')
                pass
            except (AttributeError, ProgrammingError) as e:
                raise Exception(f'Fail connecting to SQLite file: {self.db_name}', e)

        shutil.copy(self.db_name, f'{self.db_name}.bkp')

        cursor = conn.cursor()
        # www.sqlite.org/pragma.html
        # https://blog.devart.com/increasing-sqlite-performance.html
        cursor.execute("PRAGMA temp_store = MEMORY")
        cursor.execute("PRAGMA synchronous=OFF")
        cursor.execute("PRAGMA journal_mode=MEMORY")

        # get database constraints
        self.get_constraints(conn)

        return conn

    def create_db(self):

        conn = self.connect_to_db(check=False)

        # definindo um cursor
        cursor = conn.cursor()

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS [history] (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                package TEXT NULL,
                source INTEGER NOT NULL, 
                data TEXT NULL,
                stack_trace TEXT NULL,
                created_date datetime not null DEFAULT (datetime('now','localtime'))
            );
        """)

        conn.commit()

        #Must get the constraints
        self.get_constraints(conn)

    def insert_history(self, source: str, data: str, stack_trace: str = ''):
        from frida_fusion.config import Configuration

        try:
            self.insert_one('history',
                            package=Configuration.package,
                            source=source,
                            data=data if data is not None else '',
                            stack_trace=stack_trace if stack_trace is not None else '',
                            )

        except Exception as e:
            Color.pl('{!} {R}Error inserting history data:{O} %s{W}' % str(e))

    @classmethod
    def scrub(cls, input_string):
        """Clean an input string (to prevent SQL injection).

        Parameters
        ----------
        input_string : str

        Returns
        -------
        str
        """
        return ''.join(k for k in input_string if k.isalnum() or k in '_-')



import os
import sys
import re
import frida
import pkgutil
import importlib
import requests
import importlib.util
from pathlib import Path

from typing import TYPE_CHECKING

from .__meta__ import __version__
from .libs.logger import Logger

if TYPE_CHECKING:
    from .fusion import Fusion  # só no type checker


class ModuleLoaderError(Exception):
    pass


class ModuleBase(object):

    name = ''
    description = ''
    mod_path = ''

    def __init__(self, name, description):
        self.name = name
        self.description = description
        self.mod_path = str(Path(__file__).resolve().parent)
        pass

    def safe_name(self):
        return ModuleBase.get_safe_name(self.name)

    def start_module(self, **kwargs) -> bool:
        raise Exception('Method "start_module" is not yet implemented.')

    def js_files(self) -> list:
        return []

    def key_value_event(self,
                        script_location: "Fusion.ScriptLocation" = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:
        raise Exception('Method "key_value_event" is not yet implemented.')

    def data_event(self,
                   script_location: "Fusion.ScriptLocation" = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        raise Exception('Method "data_event" is not yet implemented.')

    @staticmethod
    def get_safe_name(name):
        name = name.replace(" ", "_").lower()
        return re.sub(r'[^a-zA-Z0-9_.-]+', '', name)

    @classmethod
    def _get_codeshare(cls, uri: str) -> dict:

        if uri is None or len(uri) <= 10:
            raise Exception("Invalid codeshare uri. Uri must be only user/project_name.")

        uri = uri.strip(" /@.")

        headers = {
            "Accept": "application/vnd.github+json",
            "User-Agent": f"Frida-fusion v{__version__}, Frida v{frida.__version__}"
        }

        try:
            resp = requests.get(f"https://codeshare.frida.re/api/project/{uri}", headers=headers, timeout=30)
            resp.raise_for_status()
            data = resp.json()
            if data is None:
                raise Exception("data is empty")

            if data.get('source', None) is None or data.get('source', '').strip(" \r\n") == "":
                raise Exception("source code is empty")

            return data

        except Exception as e:
            raise ModuleLoaderError("Error getting codeshare data") from e


class Module(object):
    modules = {}

    def __init__(self, name, description, module, qualname, class_name):
        self.name = name
        self.description = description
        self.module = module
        self.qualname = qualname
        self._class = class_name
        pass

    def safe_name(self):
        return ModuleBase.get_safe_name(self.name)

    def create_instance(self):
        return self._class()

    @classmethod
    def get_base_module(cls) -> str:
        file = Path(__file__).stem

        parent_module = f'.{cls.__module__}.'.replace(f'.{file}.', '').strip(' .')

        return '.'.join((parent_module, 'modules'))


class InternalModule(Module):
    pass


class ExternalModule(Module):
    pass


class ModuleManager:
    @classmethod
    def _safe_import_from_path(cls, path: Path, loaded_files: set):
        """
        Importa um .py arbitrário usando um nome único e registra o arquivo
        para não ser importado duas vezes.
        """
        real = path.resolve()

        try:
            if real in loaded_files:
                return

            # nome único, estável, baseado no caminho
            pseudo_name = (
                    "fusion_ext_"
                    + "_".join(real.parts).replace(":", "_").replace("\\", "_").replace("/", "_")
                    .replace(".", "_")
            )
            spec = importlib.util.spec_from_file_location(pseudo_name, real)
            if spec and spec.loader:
                mod = importlib.util.module_from_spec(spec)
                sys.modules[pseudo_name] = mod
                spec.loader.exec_module(mod)
                loaded_files.add(real)
        except Exception as ie:
            Logger.pl('\n{!} {R}Error loading external module: {G}%s{R}\n    {O} %s{W}' % (str(ie), str(real)))
            pass

    @classmethod
    def _import_via_pkgutil(cls, roots: list[Path], loaded_files: set):
        """
        Varre roots com pkgutil.walk_packages. Isso encontra
        - módulos .py no nível do root
        - pacotes (pastas com __init__.py) e seus submódulos
        NÃO entra em subpastas sem __init__.py (por isso depois complementamos).
        """
        str_roots = [str(p) for p in roots]
        for loader, modname, is_pkg in pkgutil.walk_packages(str_roots):
            try:
                mod = importlib.import_module(modname)
                mfile = getattr(mod, "__file__", None)
                if mfile:
                    loaded_files.add(Path(mfile).resolve())
            except Exception as ie:
                Logger.pl(
                    '\n{!} {R}Error loading internal module: {G}%s{R}\n    {O} %s{W}' % (
                        str(ie), str(loader.path)))
                pass

    @classmethod
    def _load_any_py_recursively(cls, root: Path, loaded_files: set):
        """
        Carrega *todo* arquivo .py sob root (rglob), incluindo subpastas sem __init__.py,
        sem duplicar o que já foi importado.
        """
        for py in root.rglob("*.py"):
            # exclui caches e similares
            if any(part in {"__pycache__"} for part in py.parts):
                continue
            cls._safe_import_from_path(py, loaded_files)

    @classmethod
    def list_modules(cls) -> dict:
        try:
            base_module = Module.get_base_module()
            modules: dict[str, Module] = {}

            # --- 1) Varredura padrão do seu pacote interno: <este_arquivo>/modules ---
            base_path = Path(__file__).resolve().parent / "modules"
            internal_mod_roots = [p for p in base_path.iterdir() if p.is_dir()]

            internal_mods = []

            # Vamos usar pkgutil para o pacote interno (mantém o comportamento)
            loaded_files: set[Path] = set()
            mods = [str(p) for p in internal_mod_roots]
            for loader, modname, is_pkg in pkgutil.walk_packages(mods):
                if not is_pkg:
                    # Reconstrói o caminho relativo para montar o import dentro do pacote base
                    mod_path = Path(getattr(loader, "path", ""))
                    try:
                        rel = mod_path.resolve().relative_to(base_path.resolve())
                        dotted = "." + ".".join(rel.parts) if rel.parts else ""
                    except Exception:
                        dotted = ""
                    importlib.import_module(f"{base_module}{dotted}.{modname}")
                    internal_mods.append(f"{base_module}{dotted}.{modname}")

            # --- 2) Varredura de caminhos externos via FUSION_MODULES ---
            env_value = os.environ.get("FUSION_MODULES", "").strip()
            if env_value:
                extra_roots = [Path(p).expanduser() for p in env_value.split(os.pathsep) if p.strip()]
                existing_roots = [p for p in extra_roots if p.exists() and p.is_dir()]

                # Para que pkgutil encontre módulos top-level nesses roots
                # (sem precisar de nomes de pacote), colocamos cada root no sys.path
                # durante a varredura. Usamos um conjunto para restaurar depois se preferir.
                original_sys_path = list(sys.path)
                try:
                    for root in existing_roots:
                        if str(root) not in sys.path:
                            sys.path.insert(0, str(root))

                    # 2a) Encontrar módulos top-level e pacotes (com __init__.py)
                    cls._import_via_pkgutil(existing_roots, loaded_files)

                    # 2b) Complementar: carregar QUALQUER .py (inclusive subpastas sem __init__.py)
                    for root in existing_roots:
                        cls._load_any_py_recursively(root, loaded_files)
                finally:
                    # opcional: restaurar sys.path (seguro para evitar vazamentos)
                    sys.path[:] = original_sys_path

            # --- 3) Instanciar subclasses de ModuleBase e montar o registry ---
            for i_class in ModuleBase.__subclasses__():
                t = i_class()
                key = t.safe_name()
                if key in modules:
                    raise ModuleLoaderError(
                        f"Duplicated Module name: {i_class.__module__}.{i_class.__qualname__}"
                    )

                if str(i_class.__module__) in internal_mods:
                    modules[key] = InternalModule(
                        name=t.name,
                        description=t.description,
                        module=str(i_class.__module__),
                        qualname=str(i_class.__qualname__),
                        class_name=i_class,
                    )
                else:
                    modules[key] = ExternalModule(
                        name=t.name,
                        description=t.description,
                        module=str(i_class.__module__),
                        qualname=str(i_class.__qualname__),
                        class_name=i_class,
                    )

            return modules

        except Exception as e:
            # Envolve a exceção original para manter contexto
            raise ModuleLoaderError("Error listing modules") from e



__version__ = '0.1.11'
__title__ = "Frida Fusion"
__description__ = "📱 frida-fusion - runtime mobile exploration"
__url__ = "https://github.com/helviojunior/frida-fusion"
__build__ = 0xc8879a3
__author__ = "Helvio Junior (M4v3r1ck)"
__author_email__ = "helvio_junior@hotmail.com"
__license__ = "GPL-3.0"
__copyright__ = "Copyright Helvio Junior"



#!/usr/bin/python3
# -*- coding: UTF-8 -*-
import os
import errno
import sys
import signal
from pathlib import Path

from .module import Module, ModuleManager, InternalModule, ExternalModule
from .libs.color import Color
from .libs.logger import Logger
from .__meta__ import __version__


class Configuration(object):
    ''' Stores configuration variables and functions for Frida Fusion. '''
    version = '0.0.0'

    initialized = False  # Flag indicating config has been initialized
    debug_level = 0
    cmd_line = ''
    base_path = str(Path(__file__).resolve().parent)
    db_path = os.path.join(str(Path(".").resolve()), "fusion.db")
    enabled_modules = {}

    # Device vars
    device_id = None
    use_usb = False
    remote_host = None

    # App vars
    package = ''
    pid = 0

    # Scripts
    frida_scripts = None

    # General
    out_file = None
    print_timestamp = False
    use_delay = False

    @staticmethod
    def initialize():
        '''
            Sets up default initial configuration values.
            Also sets config values based on command-line arguments.
        '''

        # Only initialize this class once
        if Configuration.initialized:
            return

        Configuration.debug_level = 0  # Verbosity level.

        # Overwrite config values with arguments (if defined)
        Configuration.load_from_arguments()

        Configuration.initialized = True

    @staticmethod
    def load_from_arguments():
        ''' Sets configuration values based on Argument.args object '''
        from .args import Arguments

        config_check = 0

        sys.argv[0] = 'frida-fusion'

        list_modules = any(['--list-modules' == word for word in sys.argv])
        #show_help = any(['-h' == word for word in sys.argv])

        if list_modules:
            mods = ModuleManager.list_modules()

            if len(mods) == 0:
                Color.pl('{!} {R}error: no modules found{R}{W}\r\n')
                sys.exit(1)

            max_name = max(iter([
                len(m.name) + 3
                for _, m in mods.items()
            ] + [15]))
            Color.pl(f"Available internal modules")
            for m in [m for _, m in mods.items() if isinstance(m, InternalModule)]:
                Color.pl(f"  {m.safe_name().ljust(max_name)} : {m.description}")

            Color.pl(f"\nAvailable external modules")
            ext_mods = [m for _, m in mods.items() if isinstance(m, ExternalModule)]
            if len(ext_mods) == 0:
                Color.pl(("  No external modules available. You can set {G}FUSION_MODULES{W} environment variable "
                          "to set an external modules Path"))
            for m in ext_mods:
                Color.pl(f"  {m.safe_name().ljust(max_name)} : {m.description}")

            print("")
            sys.exit(0)

        args = Arguments().args

        Color.pl('{+} {W}Startup parameters')

        if args.out_file is not None and args.out_file != '':
            Configuration.out_file = args.out_file
            try:
                with open(Configuration.out_file, 'a') as f:
                    # file opened for writing. write to it here
                    Logger.out_file = Configuration.out_file
                    f.write(Color.sc(Configuration.get_banner()) + '\n')
                    f.write(Color.sc('{+} {W}Startup parameters') + '\n')
                    pass
            except IOError as x:
                if x.errno == errno.EACCES:
                    Color.pl('{!} {R}error: could not open output file to write {O}permission denied{R}{W}\r\n')
                    sys.exit(1)
                elif x.errno == errno.EISDIR:
                    Color.pl('{!} {R}error: could not open output file to write {O}it is an directory{R}{W}\r\n')
                    sys.exit(1)
                else:
                    Color.pl('{!} {R}error: could not open output file to write{W}\r\n')
                    sys.exit(1)

        if args.app_id is None and args.pid == 0:
            Color.pl('{!} {R}error: you must specify either {O}--package{R} or {O}--attach-pid{R}{W}\r\n')
            Configuration.mandatory()

        if args.app_id is not None and args.pid > 0:
            Color.pl('{!} {R}error: you must specify just one parameter {O}--package{R} or {O}--attach-pid{R}{W}\r\n')
            Configuration.mandatory()

        if args.app_id is not None:
            Configuration.package = args.app_id
            Logger.pl('     {C}package:{O} %s{W}' % Configuration.package)
        elif args.pid > 0:
            Configuration.pid = args.pid
            Logger.pl('     {C}process id:{O} %s{W}' % Configuration.pid)

        if args.use_usb is True:
            Configuration.use_usb = True
            Logger.pl('     {C}device:{O} USB{W}')
        elif args.device_id is not None:
            Configuration.device_id = args.device_id
            Logger.pl('     {C}device:{O} %s{W}' % Configuration.device_id)
        elif args.use_remote and args.remote_host is not None:
            Configuration.remote_host = args.remote_host
            Logger.pl('     {C}device:{O} remote %s{W}' % Configuration.remote_host)

        if Configuration.use_usb is False and Configuration.device_id is None and Configuration.remote_host is None:
            Color.pl('{!} {R}error: you must specify just one parameter {O}--usb{R}, {O}--device{R} or {O}--remote{R}{W}\r\n')
            Configuration.mandatory()

        if args.frida_scripts is not None and args.frida_scripts.strip() != "":
            if not os.path.exists(args.frida_scripts):
                Color.pl(
                    '{!} {R}error: scripts path not found{R}{W}\r\n')
                Configuration.mandatory()
            Configuration.frida_scripts = args.frida_scripts

        if Configuration.frida_scripts is None:
            Color.pl(
                '{!} {R}error: you must specify scripts path {O}--script-path{R}{W}\r\n')
            Configuration.mandatory()

        Logger.pl('     {C}scripts path:{O} %s{W}' % Configuration.frida_scripts)

        if args.delay:
            Configuration.use_delay = True

        if args.show_time:
            Configuration.print_timestamp = True
        Logger.print_timestamp = Configuration.print_timestamp

        if str(args.debug_level).upper() not in Logger.level_map.keys():
            Color.pl(
                '{!} {R}error: invalid debug level{R}{W}\r\n')
            Configuration.mandatory()

        Configuration.debug_level = Logger.level_map.get(str(args.debug_level).upper(), 0)
        Logger.debug_level = Configuration.debug_level

        Logger.pl('     {C}min debug level:{O} %s{W}' % str(args.debug_level).upper())

        if args.enabled_modules is not None and isinstance(args.enabled_modules, list):
            mods = ModuleManager.list_modules()
            for mod in args.enabled_modules:
                fm = next(iter([
                    m
                    for _, m in mods.items()
                    if m.safe_name() == mod.lower()
                ]), None)
                if fm is None:
                    Color.pl(
                        '{!} {R}error: module {O}%s{R} not found{W}\r\n' % mod)
                    sys.exit(1)

                name = fm.safe_name()
                if name not in Configuration.enabled_modules.keys():
                    Configuration.enabled_modules[name] = fm

        if len(Configuration.enabled_modules) > 0:
            Logger.pl('     {C}modules:{O} %s{W}' % ', '.join([
                m.name
                for _, m in Configuration.enabled_modules.items()
            ]))

        Logger.pl("")

    @staticmethod
    def get_banner():
        Configuration.version = str(__version__)

        return '''
 {P}[ {R}FRIDA {P}]{G}—o—( {O}FUSION {G})—o—{P}[ {C}MOBILE TESTS {P}] {W}{D}// v%s{W}
     {O}> {W}{D}{O}hook your mobile tests with Frida{W}
    
    ''' % Configuration.version

    @staticmethod
    def mandatory():
        Color.pl(
            '{!} {R}error: missing a mandatory option, use {O}-h{R} help{W}\r\n')
        sys.exit(1)

    @staticmethod
    def exit_gracefully(code=0):
        ''' Deletes temp and exist with the given code '''
        exit(code)

    @staticmethod
    def kill(code=0):
        ''' Deletes temp and exist with the given code '''
        os.kill(os.getpid(), signal.SIGTERM)



import errno
import os.path
from pathlib import Path
from frida_fusion.libs.logger import Logger
from frida_fusion.libs.scriptlocation import ScriptLocation
from frida_fusion.module import ModuleBase


class TlsUnpinning(ModuleBase):

    def __init__(self):
        super().__init__('Multiple unpinning', 'Use frida_multiple_unpinning by Maurizio Siddu (@akabe1)')
        self.mod_path = str(Path(__file__).resolve().parent)
        self.js_file = os.path.join(self.mod_path, "frida_multiple_unpinning.js")

    def start_module(self, **kwargs) -> bool:
        if not os.path.isfile(self.js_file):
            Logger.pl("{+} Downloading CodeShare script from @akabe1/frida-multiple-unpinning")
            data = self._get_codeshare("@akabe1/frida-multiple-unpinning/")
            if data.get('source', None) is None or data.get('source', '').strip(" \r\n") == "":
                raise Exception("source code is empty")

            try:
                with open(self.js_file, "w", encoding='utf-8') as f:
                    f.write(data.get('source', ''))
            except IOError as x:
                if x.errno == errno.EACCES:
                    Logger.pl('{!} {R}error: could not open output file to write {O}permission denied{R}{W}\r\n')
                elif x.errno == errno.EISDIR:
                    Logger.pl('{!} {R}error: could not open output file to write {O}it is an directory{R}{W}\r\n')
                else:
                    Logger.pl('{!} {R}error: could not open output file to write{W}\r\n')

                # Try to save locally
                self.js_file = str(Path("frida_multiple_unpinning.js").resolve().absolute())
                with open(self.js_file, "w", encoding='utf-8') as f:
                    f.write(data.get('source', ''))

        return True

    def js_files(self) -> list:
        return [
            self.js_file
        ]

    def key_value_event(self,
                        script_location: ScriptLocation = None,
                        stack_trace: str = None,
                        module: str = None,
                        received_data: dict = None
                        ) -> bool:
        return True

    def data_event(self,
                   script_location: ScriptLocation = None,
                   stack_trace: str = None,
                   received_data: str = None
                   ) -> bool:
        return True


